// (c) Peter Kankowski, 2006 http://smallcode.weblogs.us
#include <windows.h>
#include <stdio.h>
#include <string.h>

#define NUM_OF(A) (sizeof(A)/sizeof((A)[0])) // Number of elements in array

void* memchr_naive(void* mem, int ch, size_t len) {
	char* s = (char*)mem;
	const char* end = (const char*)mem + len;
	for(; s < end; s++)
		if(*s == ch)
			return s;
	return NULL;
}

// Note: the string cannot be "const" for sentinel !
void* memchr_sentinel(void* mem, int ch, size_t len) {
	char* s = (char*)mem;
	char* sentinel = (char*)(s + len);
	*sentinel = (char)ch;
	while(*s != ch)
		s++;
	*sentinel = '\0';
	return (s < sentinel) ? s : NULL;
}

void* memchr_unroll(void* mem, int ch, size_t len) {
	char* s = (char*)mem;
	char* sentinel = (char*)(s + len);
	*sentinel = (char)ch;
	DWORD mask4 = (BYTE(ch) << 8) | BYTE(ch);
	mask4 |= (mask4 << 16);
	for(;;) {
		DWORD x = *(DWORD*)s ^ mask4;
		if((x & 0x000000FF) == 0) goto S1;
		if((x & 0x0000FF00) == 0) goto S2;
		if((x & 0x00FF0000) == 0) goto S3;
		if((x & 0xFF000000) == 0) goto S4;
		s += 4;
	}
S4:	s++;
S3:	s++;
S2:	s++;
S1:
	*sentinel = '\0';
	return (s < sentinel) ? s : NULL;
}

void* memchr_vector(void* mem, int ch, size_t len) {
	char* s = (char*)mem;
	char* sentinel = (char*)(s + len);
	*sentinel = (char)ch;
	DWORD mask4 = (BYTE(ch) << 8) | BYTE(ch);
	mask4 |= (mask4 << 16);
	DWORD y;
	do {
		DWORD x = *(DWORD*)s ^ mask4;
		y = (x - 0x01010101) & ~x;
		s += 4;
	} while( (y & 0x80808080) == 0 );
	s -= 4;
	if((y & 0x8080) == 0) {
		s += 2;
		y >>= 16;
	}
	if((y & 0x80) == 0)
		s++;
	*sentinel = '\0';
	return (s < sentinel) ? s : NULL;
}

// Vector method from Agner Fog's manual + sentinel
void* memchr_vector_asm(void* mem, int cha, size_t len) {
	__asm {
		mov eax, [mem]
		mov ecx, [len]
		mov edx, [cha]
		add ecx, eax
		
		mov ebx, edx
		shl ebx, 8
		or edx, ebx

		mov BYTE PTR [ecx], dl
		mov ebx, edx
		shl ebx, 16
		or edx, ebx
		
L:
		mov ebx, [eax]
		xor ebx, edx
		add eax, 4
		lea esi, [ebx-0x01010101]
		not ebx
		and esi, ebx
		and esi, 0x80808080
		jz L

		mov byte ptr[ecx], 0

		mov ebx, esi
		shr esi, 16
		test ebx, 0x8080
		cmovz ebx, esi
		lea esi, [eax+2]
		cmovz eax, esi
		
		add bl, bl
		sbb eax, 3

		cmp	eax, ecx // return (s < sentinel) ? s : NULL;
		sbb ebx, ebx
		and eax, ebx
	}
}



// Test data
char a1[]="Four",a2[]="score",a3[]="and",a4[]="seven",a5[]="years",a6[]="ago",a7[]="our",
a8[]="fathers",a9[]="brought",a10[]="forth",a11[]="upon",a12[]="this",
a13[]="continent",a14[]="a",a15[]="new",a16[]="nation",a17[]="conceived",
a18[]="in",a19[]="liberty",a20[]="and",a21[]="dedicated",a22[]="to",a23[]="the",
a24[]="proposition",a25[]="that",a26[]="all",a27[]="men",a28[]="are",a29[]="created",
a30[]="equal";
char *txt[] = {
	a1, a2, a3, a4, a5, a6, a7, a8, a9, a10,
	a11, a12, a13, a14, a15, a16, a17, a18, a19, a20,
	a21, a22, a23, a24, a25, a26, a27, a28, a29, a30};

int length[] = {4, 5, 3, 5, 5, 3, 3, 7, 7, 5, 4, 4, 9, 1, 3, 6, 9,
 2, 7, 3, 9, 2, 3, 11, 4, 3, 3, 3, 7, 5, };

char jabberwocky[] =
"'Twas brillig, and the slithy toves\n"
"Did gyre and gimble in the wabe:\n"
"All mimsy were the borogoves,\n"
"And the mome raths outgrabe.\n"

"Beware the Jabberwock, my son!\n"
"The jaws that bite, the claws that catch!\n"
"Beware the Jubjub bird, and shun\n"
"The frumious Bandersnatch!\n"

"He took his vorpal sword in hand:\n"
"Long time the manxome foe he sought -\n"
"So rested he by the Tumturn tree,\n"
"And stood awhile in thought.\n"

"And, as in uffish thought he stood;\n"
"The Jabberwock, with eyes of flame,\n"
"Came whiffling through the tulgey wood,\n"
"And burbled as it came!\n"

"One, two! One, two! And through and through\n"
"The vorpal blade went snicker-snackl\n"
"He left it dead, and with its head\n"
"He went galumphing back.\n"

"And hast thou slain the Jabberwock?\n"
"Come to my arms, my beamish boy!\n"
"O frabjous day! Callooh! Callay!\n"
"He chortled in his joy.\n"

"'Twas brillig, and the slithy toves\n"
"Did gyre and gimble in the wabe:\n"
"All mimsy were the borogoves,\n"
"And the mome raths outgrabe.\n\n"
"Humpty-Dumpty Magazine.";

unsigned __int64 GetRDTSC() {
	__asm {
		; Flush the pipeline
		XOR eax, eax
		CPUID
		; Get RDTSC counter in edx:eax
		RDTSC
	}
}

typedef void* (*MEMCHR)(void *s, int ch, size_t length);
typedef UINT(*TESTFUNC)(MEMCHR memchrfunc);

UINT_PTR __forceinline Test1(MEMCHR func) {
	UINT_PTR sanity_check_sum = 0;
	for(int i = 0; i < NUM_OF(txt); ++i)
		sanity_check_sum += (UINT_PTR)func(txt[i], 't', length[i]);
	return sanity_check_sum;
}

UINT_PTR __forceinline Test2(MEMCHR func) {
	return (UINT_PTR)func(jabberwocky, 'z', sizeof(jabberwocky) - 1);
}

// Profiling function
template <TESTFUNC test>
void TestMemChr(const char* label, MEMCHR func) {
	UINT_PTR sanity_check_sum = 0;
	UINT64 min_time = ~0UI64;
	printf("%14s: ", label);
	for(int j = 0; j < 5; ++j) {
		UINT64 start = GetRDTSC();
		sanity_check_sum = test(func);
		UINT64 time = GetRDTSC() - start;
		printf("%6I64d", time);
		if(time < min_time)
			min_time = (INT)time;
	}
	printf("|%6I64d ticks, sum=%8X\n", (min_time + 5) / 10 * 10, sanity_check_sum);
}

int __cdecl main(void) {
	HANDLE hProcess = GetCurrentProcess();
	HANDLE hThread = GetCurrentThread();

	DWORD sys_affinity, proc_affinity;
	GetProcessAffinityMask(hProcess, &proc_affinity, &sys_affinity);
	DWORD first_processor = sys_affinity & (-(INT)sys_affinity);
	SetProcessAffinityMask(hProcess, first_processor);

	SetPriorityClass(hProcess, REALTIME_PRIORITY_CLASS);
	SetThreadPriority(hThread, THREAD_PRIORITY_TIME_CRITICAL);

	printf("memchr - short strings\n");
	TestMemChr<Test1>("CRT", memchr);
	TestMemChr<Test1>("naive", memchr_naive);
	TestMemChr<Test1>("sentinel", memchr_sentinel);
	TestMemChr<Test1>("sent & unroll", memchr_unroll);
	TestMemChr<Test1>("sent & vector", memchr_vector);
	TestMemChr<Test1>("vector asm", memchr_vector_asm);

	printf("\nmemchr - long string\n");
	TestMemChr<Test2>("CRT", memchr);
	TestMemChr<Test2>("naive", memchr_naive);
	TestMemChr<Test2>("sentinel", memchr_sentinel);
	TestMemChr<Test2>("sent & unroll", memchr_unroll);
	TestMemChr<Test2>("sent & vector", memchr_vector);
	TestMemChr<Test2>("vector asm", memchr_vector_asm);

	SetPriorityClass(hProcess, NORMAL_PRIORITY_CLASS);
	SetThreadPriority(hThread, THREAD_PRIORITY_NORMAL);
	return 0;
}
